-----------------------
-- Game Board handling
-----------------------

BloxBoard = { }
local IsMoveOk
local fmod = math.fmod

-- Just returns an array representing an empty game board.
local function Blox_EmptyBoard()
   local board = { }
   for y = 1, BloxGameBoard.HEIGHT do
      board[y] = { }
      for x = 1, BloxGameBoard.WIDTH do
	 board[y][x] = 0
      end
   end
   return board
end

local function Blox_InitAnims(x, y)
   local anims = { }
   for y = 1, y do
      anims[y] = { }
      for x = 1, x do
	 anims[y][x] = AnimBlock:new()
      end
   end
   return anims
end

-- Create a new game board
function BloxBoard:new(o)
   if not o then
      o = {
	 shape = BloxShape:new(),
	 next = BloxShape:new(),
	 refresh_score = true,
	 refresh_board = true, 
	 board = Blox_EmptyBoard(),
	 anims = Blox_InitAnims(BloxGameBoard.WIDTH, BloxGameBoard.HEIGHT), -- game board anims
	 explode_anims = Blox_InitAnims(BloxGameBoard.WIDTH, 4), -- exploding blocks...
	 isAnimating = false,
	 isExploding = false, 
	 isPaused = true,
	 isGameOver = false,
      }
   end
   o.shape:fixTextures()
   o.next:fixTextures(true)
   setmetatable(o, self)
   if self.__index ~= self then self.__index = self end
   return o
end

-- Set the texture of a tile
function BloxBoard:SetTile(x, y, id)
   self.board[y][x] = id

   BloxBoardWidgets[x][y].backdrop.bgFile = BloxConstants.TEXTURES[id]
   BloxBoardWidgets[x][y]:SetBackdrop(BloxBoardWidgets[x][y].backdrop)
   self:AdjustBlock(BloxBoardWidgets[x][y], x, y, true, false)
end

function BloxBoard:IsTileEmpty(x, y)
   return self.board[y][x] == 0
end

-- This will move the block to the specified position
-- in the game board grid and either show or hide it.
function BloxBoard:AdjustBlock(frame, x, y, vis, next)
   local width = frame:GetWidth()
   if next then
      frame:SetPoint("TOPLEFT", x*width, -(y*width))
   else
      frame:SetPoint("BOTTOMLEFT", (x-1)*width, BloxGameBoard.HEIGHT*width - y*width)
   end
   if vis then frame:Show() else frame:Hide() end
end

function BloxBoard.GetPixelPosition(x, y) 
   local size = BloxShapeWidgets[1][1]:GetWidth()
   return (x-1)*size, BloxGameBoard.HEIGHT*size - y*size
end

-- Draw the current and next shape
function BloxBoard:DrawShapes() 
   -- Draw current shape
   local shape = self.shape
   local next = self.next
   local isMasked = shape.isMasked
   local AdjustBlock = self.AdjustBlock
   if shape.moved then
      local shouldDrawShape = IsMoveOk and not self.isPaused and not self.isGameOver
      for x = 1, 4 do
	 for y = 1, 4 do
	    local mx = x + shape.x - 1
	    local my = y + shape.y - 1
	    if shouldDrawShape then
	       AdjustBlock(self, BloxShapeWidgets[x][y], mx, my, isMasked(shape, x, y))
	    else
	       AdjustBlock(self, BloxShapeWidgets[x][y], mx, my, false)
	    end
	 end
      end
      if IsMoveOk then
	 shape.moved = false
      end
   end
   if next.moved then
      next:randomize()
      local info = next:shapeInfo()
      local xoff = (4-info[2])/2 - info[4] - 1
      local yoff = (4-info[1])/2 - info[3] - 1
      for x = 1, 4 do
	 for y = 1, 4 do
	    AdjustBlock(self, BloxNextShapeWidgets[x][y], xoff+x, yoff+y,
			isMasked(next, x, y), next)
	 end
      end
      next.moved = false
   end
end

function BloxBoard:UpdateShapes()
   self.shape:moveHorizontal()
   if not self.isAnimating then 
      while self.shape:dropTime() do
	 if not self.shape:moveDown() then
	    self.shape:drop(true)
	    self:CheckLines()
	 end
      end
   else
      self.shape.next_drop = BloxFrame.currentTicks + self.shape.drop_tick
   end
end

--------------------------------------------------------------------
-- Redraw the board, making sure all tiles are where they should be.
-- Typically only needed after lines are removed.
--------------------------------------------------------------------
function BloxBoard:RedrawBoard(redraw)
   local isPaused = self.isPaused
   local isGameOver = self.isGameOver
   local AdjustBlock = self.AdjustBlock
   for x = 1, BloxGameBoard.WIDTH do
      for y = 1, BloxGameBoard.HEIGHT do
	 local id
	 if isPaused then
	    id = BloxConstants.PAUSED[y][x]
	 elseif isGameOver then
	    id = BloxConstants.GAME_OVER[y][x]
	 else
	    id = self.board[y][x]
	 end
	 if id > 0 then
	    BloxBoardWidgets[x][y].backdrop.bgFile = BloxConstants.TEXTURES[id]
	    BloxBoardWidgets[x][y]:SetBackdrop(BloxBoardWidgets[x][y].backdrop);
	 end
	 -- Reposition, hide/show block.
	 AdjustBlock(self, BloxBoardWidgets[x][y], x, y, id > 0)
      end
   end
end


--------------------------------------------------------------------
-- Redraw the score, rows and level displays if needed.
--------------------------------------------------------------------
function BloxBoard:DrawScore()
   GameScore:SetText(BloxGame.score)
   GameLevel:SetText(BloxGame.level)
   GameRows:SetText(BloxGame.rows)
end

--------------------------------------------------------------------
-- Update the state on all moving items ss needed.
--------------------------------------------------------------------
function BloxBoard:Update()
   IsMoveOk = not self.isAnimating
   self:UpdateShapes()

   local isExploding = false
   if self.isExploding then
      for x = 1, BloxGameBoard.WIDTH do
	 for y = 1, 4 do
	    if self.explode_anims[y][x]:Update() then 
	       isExploding = true
	    end
	 end
      end
   end
   self.isExploding = isExploding
      
   local isAnimating = false
   if self.isAnimating then
      for x = 1, BloxGameBoard.WIDTH do
	 for y = 1, BloxGameBoard.HEIGHT do
	    if self.anims[y][x]:Update() then 
	       isAnimating = true
	    end
	 end
      end
   end
   self.isAnimating = isAnimating
end

--------------------------------------------------------------------
-- Redraw everything needing redrawing
--------------------------------------------------------------------
function BloxBoard:Draw()
   --   self:DrawBoard()
   self:DrawShapes()
   --   self:DrawScore()
end

--------------------------------------------------------------------
-- Check to see if there are any lines hat need to be removed.
--------------------------------------------------------------------
function BloxBoard:CheckLines()
   local removed = 0
   local cnt=0
   if self.isPaused or self.isGameOver then return end
   
   BloxGame.score = BloxGame.score + 9*BloxGame.level
   for y = 1, BloxGameBoard.HEIGHT do
      local full_line = true
      for x = 1, BloxGameBoard.WIDTH do 
	 if not BloxBoardWidgets[x][y]:IsVisible() then
	    full_line = false
	    break
	 end
      end
      if full_line then
 	 removed = removed + 1

	 self.shape.moved = true
	 for x = 1, BloxGameBoard.WIDTH do
	    AnimBlock.Create(self, self.explode_anims[removed][x], x, y, x, removed, BloxConstants.MODE_EXPLODE, 0);
	    self.board[y][x] = 0
	 end

	 self:ExplodeLine(y)
	 self.isExploding = true
      end
   end
   --   message(string.format("%d lines to remove", removed))
   while removed > 0 do
      BloxGame.score = BloxGame.score + (removed*50)*BloxGame.level
      removed = removed - 1
      BloxGame.rows = BloxGame.rows + 1
      if fmod(BloxGame.rows, 5) == 0 then
	 BloxGame.level = BloxGame.level + 1
      end
   end
   self.shape:setDropTick( BloxGame.numTicksForLevel() )
   self:DrawScore()
end

function BloxBoard:TogglePaused()
   --   if self.isPaused then
   --      GravBloxAce:Print("Unpausing game")
   --   else
   --      GravBloxAce:Print("Pausing game")
   --   end
   if self.isPaused then
      -- Set the next drop to occur in 300 ms + num ticks for level
      self.shape:setDropTick( BloxGame.numTicksForLevel() + 0.3 ) 
   end
   self.isPaused = not self.isPaused
   self:RedrawBoard()
   self.shape.moved = true -- force redraw
   self:DrawShapes()
   self:DrawScore()
end


function BloxBoard:CheckGameOver()
   if self.isPaused or self.isGameOver then
      return
   end
   if not self.shape:isValidPos() then 
      self.isGameOver = true
      self.shape.moved = true
      self:DrawShapes()
      GravBloxAce:CancelScheduledEvent("Blox_OnUpdate")
      GravBloxAce.isGrabbed = false
      self:RedrawBoard()
      BloxFrame:EnableKeyboard(false)
   end
end


function BloxBoard:ExplodeLine(y)
   self.isAnimating = true

   for x = 1, BloxGameBoard.WIDTH do
      --      block_x = x
      --      block_y = y
      --    blockPosToScreen(block_x, block_y)
      for yd = y, 1, -1 do
	 if yd > 1 then
	    self.board[yd][x] =  self.board[yd-1][x]
	    if self.board[yd][x] > 0 then
	       if self.anims[yd-1][x].active then
		  local temp = self.anims[yd][x]
		  self.anims[yd][x] = self.anims[yd-1][x]
		  self.anims[yd-1][x] = temp
		  temp.active = false
		  self.anims[yd][x].dest_x, self.anims[yd][x].dest_y = self.GetPixelPosition(x, yd)
		  self.anims[yd][x].dest_x_tile = x
		  self.anims[yd][x].dest_y_tile = yd
	       else 
		  AnimBlock.Create(self, self.anims[yd][x], x, yd-1, x, yd, BloxConstants.MODE_FALLING,
				   (BloxGameBoard.WIDTH-x)*0.03)
	       end
	       self:SetTile(x, yd, self.board[yd][x])
	    else
	       BloxBoardWidgets[x][yd]:Hide()
	    end
	 else
	    self.board[yd][x] = 0
	    BloxBoardWidgets[x][yd]:Hide()
	 end
      end
   end
end

